/*
Copyright (c) 2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/

package traceccgui;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.util.prefs.Preferences;
import dirk_krause.jtools.PlatformInformation;


/**
 *
 * @author krause
 */
public class TraceccController {
  
  /** Program names and options. */
  public static final String[] pao = {
    /*  0 */    "tracecc",
    /*  1 */    "-r",
    /*  2 */    "-l",
    /*  3 */    "-p",
    /*  4 */    "-b",
    /*  5 */    "-d",
    /*  6 */    "-s",
    /*  7 */    "-i",
    /*  8 */    "-k",
    /*  9 */    "-t",
    /* 10 */    "--/ui/lang=en_US",
    /* 11 */    "--/ui/lang/env=no",
    /* 12 */    "--/log/stderr/level=none",
    /* 13 */    "--/log/stdout/level=warning",
    /* 14 */    "--/log/file/level=none",
    /* 15 */	"jlayout",
  };
  
  /** data set for platform and user information */
  protected PlatformInformation pi = null;

  /** Preferences node */
  private String prefNode = "/TraceccGUI";

  /** Boolean values */
  protected boolean[] bv = {
    true,	//  0: always show log
    false,	//  1: check modification timestamp
    true,	//  2: ctr -> c
    true,	//  3: cpt -> cpp
    true,	//  4: mtr -> m
    true,	//  5: jtr -> java
    true,	//  6: jl  -> java
    true,	//  7: show line numbers
    true,	//  8: allow C++ comments in C files
    true,	//  9: debug
    false,	// 10: debug to stdout
    false,	// 11: debug output for ide
    false,	// 12: timestamp in debug output
    false,	// 13: trace keyword in debug output
    false,	// 14: have jlayout program
  };

  /** Boolean values preferences keys */
  protected String[] bk = {
    "show-log",		//  0 always show log
    "make",		//  1 check modification time stamp
    "suffix.ctr",	//  2 ctr
    "suffix.cpt",	//  3 cpt
    "suffix.mtr",	//  4 mtr
    "suffix.jtr",	//  5 jtr
    "suffix.jl",	//  6 jl
    "output.lineno",	//  7 line numbers
    "output.cplusplus",	//  8 C++ style comments in C
    "debug.use",	//  9 debug
    "debug.stdout",	// 10 debug to stdout
    "debug.ide.use",	// 11 prepare debug output for ide
    "debug.timestamp",	// 12 timestamp in debug output
    "debug.trace",	// 13 trace keyword in debug output
    "have-jlayout",	// 14 have jlayout, overwritten by initialization
  };

  /** Integer values */
  protected int[] iv = {
    75,			//  0: comment box width
     0,			//  1: ide type
  };

  /** Integer values preferences keys */
  protected String[] ik = {
    "output.comment-box",	//  0
    "debug.ide.style",		//  1
  };

  /** String values */
  protected String[] sv = {
    ".",			//  0: current working directory
  };

  /** String values preferences keys */
  protected String[] sk = {
    "directory.current",	//  0 current working directory
  };
  
  
  
  /** Constructor */
  public TraceccController(PlatformInformation p) {
      pi = p;
  }



  /**
   * Get a boolean value.
   *
   * @param i
   * The index in the array.
   *
   * @return
   * The value.
   */
  public synchronized boolean getBoolean(int i) {
    boolean back = false;
    if((i >= 0) && (i < bv.length)) { back = bv[i]; }
    return back;
  }



  /**
   * Get an integer value
   *
   * @param i
   * The index in the array.
   *
   * @return
   * The value.
   */
  public synchronized int getInt(int i) {
    int back = -1;
    if((i >= 0) && (i < iv.length))  { back = iv[i]; }
    return back;
  }




  /**
   * Get a string value.
   *
   * @param i
   * The index in the array.
   *
   * @return
   * The value.
   */
  public synchronized String getString(int i) {
    String back = null;
    if((i >= 0) && (i < sv.length)) { back = sv[i]; }
    return back;
  }



  /**
   * Set a new boolean value.
   *
   * @param i
   * The index of the value in the array.
   *
   * @param b
   * The new value.
   */
  public synchronized void setBoolean(int i, boolean b) {
    if((i >= 0) && (i < bv.length)) { bv[i] = b; }
  }



  /**
   * Set a new integer value
   *
   * @param i
   * The index of the value in the array.
   *
   * @param v
   * The new value.
   */
  public synchronized void setInt(int i, int v) {
    if((i >= 0) && (i < iv.length)) { iv[i] = v; }
  }



  /**
   * Set a new string value.
   *
   * @param i
   * The index of the string in the array.
   *
   * @param s
   * The new value.
   */
  public synchronized void setString(int i, String s) {
    if((i >= 0) && (i < sv.length)) { sv[i] = s; }
  }



  /**
   * Check whether or not the jlayout program is available.
   * 
   * @return
   * Flag to indicate whether or not jlayout was found.
   */
  private boolean checkForJlayout() {
    boolean back = false;
    try {
      ProcessBuilder pb = new ProcessBuilder(
        "jlayout",
	"--version",
	"--/ui/lang=en_US",
	"--/ui/lang/env=no",
	"--/log/stderr/level=none",
	"--/log/stdout/level=none",
	"--/log/file/level=none"
      );
      Process p = pb.start();
      InputStreamReader isr = new InputStreamReader(p.getInputStream());
      BufferedReader in = new BufferedReader(isr);
      for(String s; (s = in.readLine()) != null; ) {
        if(s.contains("http://jlayout.sourceforge.net")) {
	  back = true;
	}
      }
      in.close(); in = null; isr = null;
      isr = new InputStreamReader(p.getErrorStream());
      in = new BufferedReader(isr);
      for(String s; (s = in.readLine()) != null; ) { }
      in.close(); in = null; isr = null;
      p.waitFor();
    }
    catch(Exception exc) {  }
    return back;
  }



  /**
   * Attemt to restore all saved settings.
   */
  void init() {
    int i;
    sv[0] = pi.getOriginalCurrentDirectory();
    Preferences userPrefs = Preferences.userRoot().node(prefNode);
    for(i = 0; i < bv.length; i++) {
      try {
        boolean b = userPrefs.getBoolean(bk[i], bv[i]);
	bv[i] = b;
      } catch(RuntimeException e) {}
    }
    for(i = 0; i < iv.length; i++) {
      try {
        int k = userPrefs.getInt(ik[i], iv[i]);
	iv[i] = k;
      } catch(RuntimeException e) {}
    }
    for(i = 0; i < sv.length; i++) {
      try {
        String s= userPrefs.get(sk[i], sv[i]);
	sv[i] = s;
      } catch(RuntimeException e) {}
    }
    bv[14] = checkForJlayout();
    if(!bv[14]) { bv[6] = false; }
  }




  /**
   * Attempt to save all settings.
   */
  void save() {
    int i;
    Preferences userPrefs = Preferences.userRoot().node(prefNode);
    for(i = 0; i < bv.length; i++) {
      try {
        userPrefs.putBoolean(bk[i], getBoolean(i));
      } catch(RuntimeException e) {}
    }
    for(i = 0; i < iv.length; i++) {
      try {
        userPrefs.putInt(ik[i], getInt(i));
      } catch(RuntimeException e) {}
    }
    for(i = 0; i < sv.length; i++) {
      try {
        userPrefs.put(sk[i], getString(i));
      } catch(RuntimeException e) {}
    }
  }



  /**
   * Clone all attributes from a given source.
   *
   * @param ori
   * The source to clone from.
   */
  void cloneFrom(TraceccController ori) {
    int i;
    for(i = 0; i < bv.length; i++) {
      bv[i] = ori.getBoolean(i);
    }
    for(i = 0; i < iv.length; i++) {
      iv[i] = ori.getInt(i);
    }
    for(i = 0; i < sv.length; i++) {
      sv[i] = ori.getString(i);
    }
    pi = new PlatformInformation();
    pi.cloneFrom(ori.pi);
  }



}



// project tracegui
// @(#)TraceccController.java 1.2 07/21/08

