/*
Copyright (c) 2008, Dirk Krause
All rights reserved.

Redistribution and use in source and binary forms,
with or without modification, are permitted provided
that the following conditions are met:

* Redistributions of source code must retain the above
  copyright notice, this list of conditions and the
  following disclaimer.
* Redistributions in binary form must reproduce the above 
  opyright notice, this list of conditions and the following
  disclaimer in the documentation and/or other materials
  provided with the distribution.
* Neither the name of the Dirk Krause nor the names of
  its contributors may be used to endorse or promote
  products derived from this software without specific
  prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED.
IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
*/


package traceccgui;

import java.io.File;
import java.io.IOException;
import java.util.Enumeration;
import java.util.HashSet;
import dirk_krause.jtools.StringTool;


/**
 *
 * @author krause
 */
public class TraceccListOfInputFiles {
    
  /** File name suffixes to handle */
  public static final String[] suffixes = {
    ".ctr",     //  0
    ".c",       //  1
    ".cpt",     //  2
    ".cpp",     //  3
    ".mtr",     //  4
    ".m",       //  5
    ".jtr",     //  6
    ".java",    //  7
    ".jl",      //  8
  };
  
  /** The tracecc controller */
  private TraceccController tc = null;
  
  /** Sets of input files */
  protected HashSet<String>   ctrFiles = null;
  protected HashSet<String>   cptFiles = null;
  protected HashSet<String>   mtrFiles = null;
  protected HashSet<String>   jtrFiles = null;
  protected HashSet<String>   jlFiles = null;
  
  
  
  /**
   * Constructor.
   * 
   * @param t
   * The tracecc controller to use for conversions options.
   */
  protected TraceccListOfInputFiles(TraceccController t) {
    tc = t;
    ctrFiles = new HashSet<String>();
    cptFiles = new HashSet<String>();
    mtrFiles = new HashSet<String>();
    jtrFiles = new HashSet<String>();
    jlFiles = new HashSet<String>();
  }
  
  
  /**
   * Handle one file name.
   * 
   * @param fn
   * The file name to handle.
   * 
   * @param ss
   * Source file name suffix.
   * 
   * @param sd
   * Destination file name suffix.
   * 
   * @param hs
   * Hash set to enter file name.
   * 
   * @param canHandle
   * Flag to indicate whether the file suffix must be handled.
   */
  private void checkFile(
    String fn, String ss, String sd, HashSet<String> hs, boolean canHandle
  ) {
    boolean fileNameMatches = false;
    if(canHandle) {
      if(fn.length() > ss.length()) {
        if(tc.pi.isCaseSensitive()) {
          if(fn.endsWith(ss)) {
            fileNameMatches = true;
          }
        } else {
          String suffix = fn.substring((fn.length() - ss.length()));
          if(suffix.equalsIgnoreCase(ss)) {
            fileNameMatches = true;
          }
        }
        if(fileNameMatches) {
          boolean mustHandleThisFile = true;
          if(tc.bv[1]) {
            StringBuilder sb = new StringBuilder();
            sb.append(fn.substring(0, (fn.length() - ss.length())));
            sb.append(sd);
            String destName = sb.toString();
            File fs = new File(fn);
            File fd = new File(destName);
            if(fs.exists()) {
              if(fd.exists()) {
                if(fs.isFile()) {
                  if(fd.isFile()) {
                    if(fd.lastModified() > fs.lastModified()) {
                      mustHandleThisFile = false;
                    }
                  }
                }
              }
            }
          }
          if(mustHandleThisFile) {
            hs.add(fn);
          }
        }
      }
    } else {
    }
  }
  
  
  /**
   * Check for .jl suffix and handle file name.
   * 
   * @param fn
   * The file name to check.
   */
  private void checkJlFile(String fn) {
    if(fn.length() > suffixes[8].length()) {
      boolean suffixMatches = false;
      if(tc.pi.isCaseSensitive()) {
        suffixMatches = fn.endsWith(suffixes[8]);  
      } else {
        String ss = fn.substring((fn.length() - suffixes[8].length()));
        suffixMatches = ss.equalsIgnoreCase(suffixes[8]);
      }
      if(suffixMatches) {
        jlFiles.add(fn);
      }
    }
  }

  
  
  /**
   * Retrieve all source file names in the specified directory,
   * build the hash sets.
   */
  protected void getFilenames() {
    File fCurrentDirectory = new File(tc.sv[0]);
    if(fCurrentDirectory.exists()) {
      if(fCurrentDirectory.isDirectory()) {
        String allFileNames[] = fCurrentDirectory.list();
        for(int i = 0; i < allFileNames.length; i++) {
          StringBuilder sb = new StringBuilder();
          sb.append(tc.sv[0]);
          sb.append(File.separator);
          sb.append(allFileNames[i]);
          String fileName = sb.toString();
          File f = new File(fileName);
          if(f.exists()) {
            if(f.isFile()) {
              checkFile(fileName, suffixes[0], suffixes[1], ctrFiles, tc.bv[2]);
              checkFile(fileName, suffixes[2], suffixes[3], cptFiles, tc.bv[3]);
              checkFile(fileName, suffixes[4], suffixes[5], mtrFiles, tc.bv[4]);
              checkFile(fileName, suffixes[6], suffixes[7], jtrFiles, tc.bv[5]);
              checkJlFile(fileName);
            }
          }
        }
      }
    }
  }
  
}



// project tracegui
// @(#)TraceccListOfInputFiles.java 1.2 07/09/08

